from __future__ import annotations

import typing as _t
import weakref
from reprlib import recursive_repr as _recursive_repr

import typing_extensions as _te
from reprit.base import generate_repr as _generate_repr

from ._core.abcs import (NIL,
                         Nil as _Nil,
                         Node as _Node,
                         Tree as _Tree)
from ._core.hints import (Item as _Item,
                          Key as _Key,
                          Order as _Order,
                          Value as _Value)
from ._core.maps import Map as _Map
from ._core.sets import (KeyedSet as _KeyedSet,
                         Set as _Set)
from ._core.utils import (dereference_maybe as _dereference_maybe,
                          maybe_weakref as _maybe_weakref,
                          split_items as _split_items,
                          to_balanced_tree_height as _to_balanced_tree_height,
                          to_unique_sorted_items as _to_unique_sorted_items,
                          to_unique_sorted_values as _to_unique_sorted_values)


class Node(_t.Generic[_Key, _Value]):
    _left: _t.Optional[_te.Self]
    _right: _t.Optional[_te.Self]
    _parent: _t.Optional['weakref.ref[_te.Self]']

    __slots__ = ('is_black', '_key', '_left', '_parent', '_right', '_value',
                 '__weakref__')

    def __init__(self,
                 key: _Key,
                 value: _Value,
                 is_black: bool,
                 left: _t.Union[_Nil, _te.Self] = NIL,
                 right: _t.Union[_Nil, _te.Self] = NIL,
                 parent: _t.Union[_Nil, _te.Self] = NIL) -> None:
        self._key, self._value, self.is_black = key, value, is_black
        self.left, self.right, self.parent = left, right, parent

    __repr__ = _recursive_repr()(_generate_repr(__init__))

    def __getstate__(self) -> _t.Tuple[_t.Any, ...]:
        return (self._key, self.value, self.is_black,
                self.parent, self.left, self.right)

    def __setstate__(self, state: _t.Tuple[_t.Any, ...]) -> None:
        (self._key, self._value, self.is_black,
         self.parent, self._left, self._right) = state

    @classmethod
    def from_simple(cls: _t.Type[Node[_Key, _Key]],
                    key: _Key,
                    *args: _t.Any) -> Node[_Key, _Key]:
        return cls(key, key, *args)

    @property
    def item(self) -> _Item[_Key, _Value]:
        return self.key, self.value

    @property
    def key(self) -> _Key:
        return self._key

    @property
    def left(self) -> _t.Union[_Nil, _te.Self]:
        return self._left

    @left.setter
    def left(self, node: _t.Union[_Nil, _te.Self]) -> None:
        self._left = node
        _set_parent(node, self)

    @property
    def parent(self) -> _t.Union[_Nil, _te.Self]:
        return _dereference_maybe(self._parent)

    @parent.setter
    def parent(self, node: _t.Union[_Nil, _te.Self]) -> None:
        self._parent = _maybe_weakref(node)

    @property
    def right(self) -> _t.Union[_Nil, _te.Self]:
        return self._right

    @right.setter
    def right(self, node: _t.Union[_Nil, _te.Self]) -> None:
        self._right = node
        _set_parent(node, self)

    @property
    def value(self) -> _Value:
        return self._value

    @value.setter
    def value(self, value: _Value) -> None:
        self._value = value


def _set_parent(node: _t.Union[_Nil, Node[_Key, _Value]],
                parent: _t.Optional[Node[_Key, _Value]]) -> None:
    if node is not NIL:
        node.parent = parent


def _set_black(maybe_node: _t.Optional[Node[_Key, _Value]]) -> None:
    if maybe_node is not None:
        maybe_node.is_black = True


def _is_left_child(node: Node[_Key, _Value]) -> bool:
    parent = node.parent
    return parent is not None and parent.left is node


def _is_node_black(node: _t.Union[_Nil, Node[_Key, _Value]]) -> bool:
    return node is NIL or node.is_black


class Tree(_Tree[_Key, _Value]):
    root: _t.Optional[Node[_Key, _Value]]

    @staticmethod
    def predecessor(
            node: _Node[_Key, _Value]
    ) -> _t.Union[_Nil, Node[_Key, _Value]]:
        assert isinstance(node, Node)
        if node.left is NIL:
            result = node.parent
            while result is not None and node is result.left:
                node, result = result, result.parent
        else:
            result = node.left
            while result.right is not NIL:
                result = result.right
        return result

    @staticmethod
    def successor(
            node: _Node[_Key, _Value]
    ) -> _t.Union[_Nil, Node[_Key, _Value]]:
        assert isinstance(node, Node)
        if node.right is NIL:
            result = node.parent
            while result is not None and node is result.right:
                node, result = result, result.parent
        else:
            result = node.right
            while result.left is not NIL:
                result = result.left
        return result

    @_t.overload
    @classmethod
    def from_components(cls,
                        _keys: _t.Iterable[_Key],
                        _values: None = ...) -> Tree[_Key, _Key]:
        ...

    @_t.overload
    @classmethod
    def from_components(cls,
                        _keys: _t.Iterable[_Key],
                        _values: _t.Iterable[_Value]) -> _te.Self:
        ...

    @classmethod
    def from_components(
            cls: _t.Union[
                _t.Type[Tree[_Key, _Key]], _t.Type[Tree[_Key, _Value]]
            ],
            _keys: _t.Iterable[_Key],
            _values: _t.Optional[_t.Iterable[_Value]] = None
    ) -> _t.Union[Tree[_Key, _Key], Tree[_Key, _Value]]:
        keys = list(_keys)
        if not keys:
            return cls(NIL)
        elif _values is None:
            keys = _to_unique_sorted_values(keys)

            def to_simple_node(
                    start_index: int,
                    end_index: int,
                    depth: int,
                    height: int = _to_balanced_tree_height(len(keys)),
                    constructor: _t.Callable[..., Node[_Key, _Key]]
                    = Node.from_simple
            ) -> Node[_Key, _Key]:
                middle_index = (start_index + end_index) // 2
                return constructor(
                        keys[middle_index],
                        depth != height,
                        (to_simple_node(start_index, middle_index, depth + 1)
                         if middle_index > start_index
                         else NIL),
                        (to_simple_node(middle_index + 1, end_index, depth + 1)
                         if middle_index < end_index - 1
                         else NIL)
                )

            simple_root = to_simple_node(0, len(keys), 0)
            simple_root.is_black = True
            return _t.cast(_t.Type[Tree[_Key, _Key]], cls)(simple_root)
        else:
            items = _to_unique_sorted_items(keys, tuple(_values))

            def to_complex_node(
                    start_index: int,
                    end_index: int,
                    depth: int,
                    height: int = _to_balanced_tree_height(len(items)),
                    constructor: _t.Callable[..., Node[_Key, _Value]] = Node
            ) -> Node[_Key, _Value]:
                middle_index = (start_index + end_index) // 2
                return constructor(
                        *items[middle_index],
                        depth != height,
                        (to_complex_node(start_index, middle_index, depth + 1)
                         if middle_index > start_index
                         else NIL),
                        (to_complex_node(middle_index + 1, end_index,
                                         depth + 1)
                         if middle_index < end_index - 1
                         else NIL)
                )

            complex_root = to_complex_node(0, len(items), 0)
            complex_root.is_black = True
            return _t.cast(_t.Type[Tree[_Key, _Value]], cls)(complex_root)

    def insert(self, key: _Key, value: _Value) -> Node[_Key, _Value]:
        parent = self.root
        if parent is NIL:
            node = self.root = Node(key, value, True)
            return node
        while True:
            if key < parent.key:
                if parent.left is NIL:
                    node = Node(key, value, False)
                    parent.left = node
                    break
                else:
                    parent = parent.left
            elif parent.key < key:
                if parent.right is NIL:
                    node = Node(key, value, False)
                    parent.right = node
                    break
                else:
                    parent = parent.right
            else:
                return parent
        self._restore(node)
        return node

    def remove(self, node: _Node[_Key, _Value]) -> None:
        assert isinstance(node, Node)
        successor, is_node_black = node, node.is_black
        if successor.left is NIL:
            (successor_child, successor_child_parent,
             is_successor_child_left) = (successor.right, successor.parent,
                                         _is_left_child(successor))
            self._transplant(successor, successor_child)
        elif successor.right is NIL:
            (successor_child, successor_child_parent,
             is_successor_child_left) = (successor.left, successor.parent,
                                         _is_left_child(successor))
            self._transplant(successor, successor_child)
        else:
            assert node.right is not NIL
            successor = node.right
            while successor.left is not NIL:
                successor = successor.left
            is_node_black = successor.is_black
            successor_child, is_successor_child_left = successor.right, False
            if successor.parent is node:
                successor_child_parent = successor
            else:
                is_successor_child_left = _is_left_child(successor)
                successor_child_parent = successor.parent
                self._transplant(successor, successor.right)
                successor.right = node.right
            self._transplant(node, successor)
            assert node.left is not NIL
            node.left.parent = successor
            successor.left, successor.is_black = node.left, node.is_black
        if is_node_black:
            self._remove_node_fixup(successor_child, successor_child_parent,
                                    is_successor_child_left)

    def _restore(self, node: Node[_Key, _Value]) -> None:
        while not _is_node_black(node.parent):
            parent = node.parent
            assert parent is not NIL
            grandparent = parent.parent
            assert grandparent is not NIL
            if parent is grandparent.left:
                assert grandparent is not NIL
                uncle = grandparent.right
                if _is_node_black(uncle):
                    if node is parent.right:
                        self._rotate_left(parent)
                        node, parent = parent, node
                    parent.is_black, grandparent.is_black = True, False
                    assert grandparent is not NIL
                    self._rotate_right(grandparent)
                else:
                    assert uncle is not NIL
                    parent.is_black = uncle.is_black = True
                    grandparent.is_black = False
                    node = grandparent
            else:
                uncle = grandparent.left
                if _is_node_black(uncle):
                    if node is parent.left:
                        self._rotate_right(parent)
                        node, parent = parent, node
                    parent.is_black, grandparent.is_black = True, False
                    assert grandparent is not NIL
                    self._rotate_left(grandparent)
                else:
                    assert uncle is not NIL
                    parent.is_black = uncle.is_black = True
                    grandparent.is_black = False
                    node = grandparent
        assert self.root is not NIL
        self.root.is_black = True

    def _remove_node_fixup(self,
                           node: _t.Union[_Nil, Node[_Key, _Value]],
                           parent: _t.Union[_Nil, Node[_Key, _Value]],
                           is_left_child: bool) -> None:
        while node is not self.root and _is_node_black(node):
            assert parent is not NIL
            if is_left_child:
                sibling = parent.right
                assert sibling is not NIL
                if not _is_node_black(sibling):
                    sibling.is_black, parent.is_black = True, False
                    self._rotate_left(parent)
                    sibling = parent.right
                assert sibling is not NIL
                if (_is_node_black(sibling.left)
                        and _is_node_black(sibling.right)):
                    sibling.is_black = False
                    assert parent is not NIL
                    node, parent = parent, parent.parent
                    is_left_child = _is_left_child(node)
                else:
                    if _is_node_black(sibling.right):
                        assert sibling is not NIL
                        assert sibling.left is not NIL
                        sibling.left.is_black, sibling.is_black = True, False
                        self._rotate_right(sibling)
                        sibling = parent.right
                    assert sibling is not NIL
                    sibling.is_black, parent.is_black = parent.is_black, True
                    _set_black(sibling.right)
                    self._rotate_left(parent)
                    node = self.root
            else:
                sibling = parent.left
                if not _is_node_black(sibling):
                    assert sibling is not NIL
                    sibling.is_black, parent.is_black = True, False
                    self._rotate_right(parent)
                    sibling = parent.left
                assert sibling is not NIL
                if (_is_node_black(sibling.left)
                        and _is_node_black(sibling.right)):
                    assert sibling is not NIL
                    sibling.is_black = False
                    assert parent is not NIL
                    node, parent = parent, parent.parent
                    is_left_child = _is_left_child(node)
                else:
                    if _is_node_black(sibling.left):
                        assert sibling.right is not NIL
                        sibling.right.is_black, sibling.is_black = True, False
                        self._rotate_left(sibling)
                        sibling = parent.left
                    assert sibling is not NIL
                    sibling.is_black, parent.is_black = parent.is_black, True
                    _set_black(sibling.left)
                    self._rotate_right(parent)
                    node = self.root
        _set_black(node)

    def _rotate_left(self, node: Node[_Key, _Value]) -> None:
        replacement = node.right
        assert replacement is not NIL
        self._transplant(node, replacement)
        node.right, replacement.left = replacement.left, node

    def _rotate_right(self, node: Node[_Key, _Value]) -> None:
        replacement = node.left
        assert replacement is not NIL
        self._transplant(node, replacement)
        node.left, replacement.right = replacement.right, node

    def _transplant(self,
                    origin: Node[_Key, _Value],
                    replacement: _t.Union[_Nil, Node[_Key, _Value]]) -> None:
        parent = origin.parent
        if parent is None:
            self.root = replacement
            _set_parent(replacement, None)
        elif origin is parent.left:
            parent.left = replacement
        else:
            parent.right = replacement

    __slots__ = 'root',

    def __init__(self, root: _t.Union[_Nil, Node[_Key, _Value]]) -> None:
        self.root = root


def map_(*items: _Item[_Key, _Value]) -> _Map[_Key, _Value]:
    return _Map(Tree.from_components(*_split_items(items)))


@_t.overload
def set_(*values: _Value,
         key: None = ...) -> _Set[_Value]:
    ...


@_t.overload
def set_(*values: _Value,
         key: _Order[_Value, _Key]) -> _KeyedSet[_Key, _Value]:
    ...


def set_(
        *values: _Value,
        key: _t.Optional[_Order[_Value, _Key]] = None
) -> _t.Union[_KeyedSet[_Key, _Value], _Set[_Value]]:
    return (_Set(Tree.from_components(values))
            if key is None
            else _KeyedSet(Tree.from_components([key(value)
                                                 for value in values],
                                                values),
                           key))



import weakref
from collections import deque
from itertools import (count,
                       groupby)
from typing import (Any,
                    Iterable,
                    List,
                    Optional,
                    Sequence,
                    Tuple,
                    overload)

from .hints import (Item,
                    Key,
                    Value)


class AntisymmetricKeyIndex:
    __slots__ = 'key', 'index'

    def __init__(self, key_index: Tuple[Key, int]) -> None:
        self.key, self.index = key_index

    @overload
    def __eq__(self, other: 'AntisymmetricKeyIndex') -> bool:
        ...

    @overload
    def __eq__(self, other: Any) -> Any:
        ...

    def __eq__(self, other: Any) -> Any:
        return (are_keys_equal(self.key, other.key)
                if isinstance(other, AntisymmetricKeyIndex)
                else NotImplemented)


def are_keys_equal(left: Key, right: Key) -> bool:
    return not (left < right or right < left)


def capacity(iterable: Iterable[Any]) -> int:
    """
    Returns number of elements in iterable.

    >>> capacity(range(0))
    0
    >>> capacity(range(10))
    10
    """
    counter = count()
    # order matters: if `counter` goes first,
    # then it will be incremented even for empty `iterable`
    deque(zip(iterable, counter),
          maxlen=0)
    return next(counter)


def to_balanced_tree_height(size: int) -> int:
    return size.bit_length() - 1


def dereference_maybe(
        maybe_reference: Optional['weakref.ref[Value]']
) -> Optional[Value]:
    return (maybe_reference
            if maybe_reference is None
            else maybe_reference())


def maybe_weakref(
        object_: Optional[Value]
) -> Optional['weakref.ReferenceType[Value]']:
    return (object_
            if object_ is None
            else weakref.ref(object_))


def to_unique_sorted_items(
        keys: Sequence[Key], values: Sequence[Value]
) -> Sequence[Item[Key, Value]]:
    return [(index_key.key, values[-index_key.index])
            for index_key, _ in groupby(
                sorted([(key, -index) for index, key in enumerate(keys)]),
                key=AntisymmetricKeyIndex)]


def to_unique_sorted_values(values: List[Value]) -> List[Value]:
    values.sort()
    return [value for value, _ in groupby(values)]


def split_items(
        items: Sequence[Tuple[Key, Value]]
) -> Tuple[Sequence[Key], Sequence[Value]]:
    keys, values = tuple(zip(*items)) if items else ((), ())
    return keys, values



from __future__ import annotations

import typing as _t
from abc import abstractmethod

from reprit.base import generate_repr

from .abcs import (NIL,
                   MutableSet,
                   Tree)
from .hints import (Key,
                    Order,
                    Value)


class BaseSet(MutableSet[Value]):
    __slots__ = 'tree',

    def __init__(self, tree: Tree[_t.Any, Value]) -> None:
        self.tree = tree

    __repr__ = generate_repr(__init__)

    def __iter__(self) -> _t.Iterator[Value]:
        for node in self.tree:
            yield node.value

    def __len__(self) -> int:
        return len(self.tree)

    def __reversed__(self) -> _t.Iterator[Value]:
        for node in reversed(self.tree):
            yield node.value

    @abstractmethod
    def ceil(self, value: Value) -> Value:
        """Returns first value not less than the given one."""

    def clear(self) -> None:
        self.tree.clear()

    @abstractmethod
    def floor(self, value: Value) -> Value:
        """Returns first value not greater than the given one."""

    def max(self) -> Value:
        node = self.tree.max()
        if node is NIL:
            raise ValueError('Set is empty')
        return node.value

    def min(self) -> Value:
        node = self.tree.min()
        if node is NIL:
            raise ValueError('Set is empty')
        return node.value

    @abstractmethod
    def next(self, value: Value) -> Value:
        """Returns first value greater than the given one."""

    def popmax(self) -> Value:
        node = self.tree.popmax()
        if node is NIL:
            raise ValueError('Set is empty')
        return node.value

    def popmin(self) -> Value:
        node = self.tree.popmin()
        if node is NIL:
            raise ValueError('Set is empty')
        return node.value

    pop = popmin

    @abstractmethod
    def prev(self, value: Value) -> Value:
        """Returns last value lesser than the given one."""


class Set(BaseSet[Value]):
    def __contains__(self, value: Value) -> bool:
        return self.tree.find(value) is not NIL

    def __copy__(self) -> Set[Value]:
        return Set(self.tree.__copy__())

    def add(self, value: Value) -> None:
        self.tree.insert(value, value)

    def ceil(self, value: Value) -> Value:
        node = self.tree.supremum(value)
        if node is NIL:
            raise ValueError('No value found greater than or equal to {!r}'
                             .format(value))
        return node.value

    def discard(self, value: Value) -> None:
        node = self.tree.find(value)
        if node is NIL:
            return
        self.tree.remove(node)

    def floor(self, value: Value) -> Value:
        node = self.tree.infimum(value)
        if node is NIL:
            raise ValueError('No value found less than or equal to {!r}'
                             .format(value))
        return node.value

    def from_iterable(self, _value: _t.Iterable[Key]) -> Set[Key]:
        return Set(self.tree.from_components(_value))

    def next(self, value: Value) -> Value:
        node = self.tree.find(value)
        if node is NIL:
            raise ValueError('{!r} is not in set'.format(value))
        node = self.tree.successor(node)
        if node is NIL:
            raise ValueError('Corresponds to maximum')
        return node.value

    def prev(self, value: Value) -> Value:
        node = self.tree.find(value)
        if node is NIL:
            raise ValueError('{!r} is not in set'.format(value))
        node = self.tree.predecessor(node)
        if node is NIL:
            raise ValueError('Corresponds to minimum')
        return node.value

    def remove(self, value: Value) -> None:
        node = self.tree.pop(value)
        if node is NIL:
            raise ValueError('{!r} is not in set'.format(value))


class KeyedSet(_t.Generic[Key, Value], BaseSet[Value]):
    __slots__ = 'key',

    def __init__(self, tree: Tree[Key, Value], key: Order[Value, Key]) -> None:
        super().__init__(tree)
        self.key = key

    __repr__ = generate_repr(__init__)

    def __contains__(self, value: Value) -> bool:
        return bool(self.tree) and self.tree.find(self.key(value)) is not NIL

    def __copy__(self) -> KeyedSet[Key, Value]:
        return KeyedSet(self.tree.__copy__(), self.key)

    def add(self, value: Value) -> None:
        self.tree.insert(self.key(value), value)

    def ceil(self, value: Value) -> Value:
        node = self.tree.supremum(self.key(value))
        if node is NIL:
            raise ValueError('No value found greater than or equal to {!r}'
                             .format(value))
        return node.value

    def discard(self, value: Value) -> None:
        node = self.tree.find(self.key(value))
        if node is NIL:
            return
        self.tree.remove(node)

    def floor(self, value: Value) -> Value:
        node = self.tree.infimum(self.key(value))
        if node is NIL:
            raise ValueError('No value found less than or equal to {!r}'
                             .format(value))
        return node.value

    def from_iterable(self,
                      _value: _t.Iterable[Value]) -> KeyedSet[Key, Value]:
        values = list(_value)
        return KeyedSet(self.tree.from_components(map(self.key, values),
                                                  values),
                        self.key)

    def next(self, value: Value) -> Value:
        key = self.key(value)
        node = self.tree.find(key)
        if node is NIL:
            raise ValueError('{!r} is not in set'.format(value))
        node = self.tree.successor(node)
        if node is NIL:
            raise ValueError('Corresponds to maximum')
        return node.value

    def prev(self, value: Value) -> Value:
        key = self.key(value)
        node = self.tree.find(key)
        if node is NIL:
            raise ValueError('{!r} is not in set'.format(value))
        node = self.tree.predecessor(node)
        if node is NIL:
            raise ValueError('Corresponds to minimum')
        return node.value

    def remove(self, value: Value) -> None:
        node = self.tree.pop(self.key(value))
        if node is NIL:
            raise ValueError('{!r} is not in set'.format(value))


def set_constructor(
        tree_constructor: _t.Callable[..., Tree[Key, Value]],
        *values: Value,
        key: _t.Optional[Order[Value, Key]] = None
) -> BaseSet[Value]:
    return (Set(tree_constructor(values))
            if key is None
            else KeyedSet(tree_constructor([key(value) for value in values],
                                           values),
                          key))



import typing as _t

import typing_extensions as _te

_T = _t.TypeVar('_T',
                contravariant=True)


class Ordered(_te.Protocol):
    def __lt__(self, other: _te.Self) -> bool:
        ...


Key = _t.TypeVar('Key',
                 bound=Ordered)
Key_co = _t.TypeVar('Key_co',
                    bound=Ordered,
                    covariant=True)
Value = _t.TypeVar('Value',
                   bound=_t.Any)
Order = _t.Callable[[Value], Key]
Item = _t.Tuple[Key, Value]



from __future__ import annotations

import typing as _t
from reprlib import recursive_repr as _recursive_repr

import typing_extensions as _te
from reprit.base import generate_repr as _generate_repr

from ._core.abcs import (NIL,
                         Nil as _Nil,
                         Node as _Node,
                         Tree as _Tree)
from ._core.hints import (Item as _Item,
                          Key as _Key,
                          Order as _Order,
                          Value as _Value)
from ._core.maps import Map as _Map
from ._core.sets import (KeyedSet as _KeyedSet,
                         Set as _Set)
from ._core.utils import (dereference_maybe as _dereference_maybe,
                          maybe_weakref as _maybe_weakref,
                          split_items as _split_items,
                          to_unique_sorted_items as _to_unique_sorted_items,
                          to_unique_sorted_values as _to_unique_sorted_values)


class Node(_t.Generic[_Key, _Value]):
    __slots__ = ('height', '_key', '_left', '_parent', '_right', '_value',
                 '__weakref__')

    def __init__(self,
                 key: _Key,
                 value: _Value,
                 left: _t.Union[_Nil, _te.Self] = NIL,
                 right: _t.Union[_Nil, _te.Self] = NIL,
                 parent: _t.Union[_Nil, _te.Self] = NIL) -> None:
        self._key, self._value = key, value
        self.left, self.right, self.parent = left, right, parent
        self.height = max(_to_height(self.left), _to_height(self.right)) + 1

    __repr__ = _recursive_repr()(_generate_repr(__init__))

    def __getstate__(self) -> _t.Tuple[
        _Key, _Value, int, _t.Union[_Nil, _te.Self],
        _t.Union[_Nil, _te.Self], _t.Union[_Nil, _te.Self]
    ]:
        return (self._key, self._value, self.height,
                self.parent, self.left, self.right)

    def __setstate__(
            self,
            state: _t.Tuple[
                _Key, _Value, int, _t.Union[_Nil, _te.Self],
                _t.Union[_Nil, _te.Self], _t.Union[_Nil, _te.Self]
            ]
    ) -> None:
        (self._key, self._value, self.height,
         self.parent, self._left, self._right) = state

    @classmethod
    def from_simple(cls: _t.Type[Node[_Key, _Key]],
                    key: _Key,
                    *args: _t.Any) -> Node[_Key, _Key]:
        return cls(key, key, *args)

    @property
    def balance_factor(self) -> int:
        return _to_height(self.left) - _to_height(self.right)

    @property
    def item(self) -> _Item[_Key, _Value]:
        return self.key, self.value

    @property
    def key(self) -> _Key:
        return self._key

    @property
    def left(self) -> _t.Union[_Nil, _te.Self]:
        return self._left

    @left.setter
    def left(self, node: _t.Union[_Nil, _te.Self]) -> None:
        self._left = node
        _set_parent(node, self)

    @property
    def parent(self) -> _t.Union[_Nil, _te.Self]:
        return _dereference_maybe(self._parent)

    @parent.setter
    def parent(self, node: _t.Union[_Nil, _te.Self]) -> None:
        self._parent = _maybe_weakref(node)

    @property
    def right(self) -> _t.Union[_Nil, _te.Self]:
        return self._right

    @right.setter
    def right(self, node: _t.Union[_Nil, _te.Self]) -> None:
        self._right = node
        _set_parent(node, self)

    @property
    def value(self) -> _Value:
        return self._value

    @value.setter
    def value(self, value: _Value) -> None:
        self._value = value


def _to_height(node: _t.Union[_Nil, Node[_Key, _Value]]) -> int:
    return -1 if node is NIL else node.height


def _update_height(node: Node[_Key, _Value]) -> None:
    node.height = max(_to_height(node.left), _to_height(node.right)) + 1


def _set_parent(node: _t.Union[_Nil, Node[_Key, _Value]],
                parent: _t.Optional[Node[_Key, _Value]]) -> None:
    if node is not NIL:
        node.parent = parent


class Tree(_Tree[_Key, _Value]):
    root: _t.Optional[Node[_Key, _Value]]

    @staticmethod
    def predecessor(
            node: _Node[_Key, _Value]
    ) -> _t.Union[_Nil, Node[_Key, _Value]]:
        assert isinstance(node, Node)
        if node.left is NIL:
            result = node.parent
            while result is not None and node is result.left:
                node, result = result, result.parent
        else:
            result = node.left
            while result.right is not NIL:
                result = result.right
        return result

    @staticmethod
    def successor(
            node: _Node[_Key, _Value]
    ) -> _t.Union[_Nil, Node[_Key, _Value]]:
        assert isinstance(node, Node)
        if node.right is NIL:
            result = node.parent
            while result is not None and node is result.right:
                node, result = result, result.parent
        else:
            result = node.right
            while result.left is not NIL:
                result = result.left
        return result

    @_t.overload
    @classmethod
    def from_components(cls,
                        _keys: _t.Iterable[_Key],
                        _values: None = ...) -> Tree[_Key, _Key]:
        ...

    @_t.overload
    @classmethod
    def from_components(cls,
                        _keys: _t.Iterable[_Key],
                        _values: _t.Iterable[_Value]) -> Tree[_Key, _Value]:
        ...

    @classmethod
    def from_components(
            cls: _t.Union[
                _t.Type[Tree[_Key, _Key]], _t.Type[Tree[_Key, _Value]]
            ],
            _keys: _t.Iterable[_Key],
            _values: _t.Optional[_t.Iterable[_Value]] = None
    ) -> _t.Union[Tree[_Key, _Key], Tree[_Key, _Value]]:
        keys = list(_keys)
        if not keys:
            return cls(NIL)
        elif _values is None:
            keys = _to_unique_sorted_values(keys)

            def to_simple_node(
                    start_index: int,
                    end_index: int,
                    constructor: _t.Callable[..., Node[_Key, _Key]]
                    = Node.from_simple
            ) -> Node[_Key, _Key]:
                middle_index = (start_index + end_index) // 2
                return constructor(keys[middle_index],
                                   (to_simple_node(start_index, middle_index)
                                    if middle_index > start_index
                                    else NIL),
                                   (to_simple_node(middle_index + 1, end_index)
                                    if middle_index < end_index - 1
                                    else NIL))

            return _t.cast(_t.Type[Tree[_Key, _Key]], cls)(
                    to_simple_node(0, len(keys))
            )
        else:
            items = _to_unique_sorted_items(keys, list(_values))

            def to_complex_node(
                    start_index: int,
                    end_index: int,
                    constructor: _t.Callable[..., Node[_Key, _Value]]
                    = Node[_Key, _Value]
            ) -> Node[_Key, _Value]:
                middle_index = (start_index + end_index) // 2
                return constructor(*items[middle_index],
                                   (to_complex_node(start_index, middle_index)
                                    if middle_index > start_index
                                    else NIL),
                                   (to_complex_node(middle_index + 1,
                                                    end_index)
                                    if middle_index < end_index - 1
                                    else NIL))

            return _t.cast(_t.Type[Tree[_Key, _Value]], cls)(
                    to_complex_node(0, len(items))
            )

    def insert(self, key: _Key, value: _Value) -> Node[_Key, _Value]:
        parent = self.root
        if parent is NIL:
            node = self.root = Node(key, value)
            return node
        while True:
            if key < parent.key:
                if parent.left is NIL:
                    node = Node(key, value)
                    parent.left = node
                    break
                else:
                    parent = parent.left
            elif parent.key < key:
                if parent.right is NIL:
                    node = Node(key, value)
                    parent.right = node
                    break
                else:
                    parent = parent.right
            else:
                return parent
        self._rebalance(node.parent)
        return node

    def remove(self, node: _Node[_Key, _Value]) -> None:
        assert isinstance(node, Node)
        if node.left is NIL:
            imbalanced_node = node.parent
            self._transplant(node, node.right)
        elif node.right is NIL:
            imbalanced_node = node.parent
            self._transplant(node, node.left)
        else:
            successor = node.right
            while successor.left is not NIL:
                successor = successor.left
            if successor.parent is node:
                imbalanced_node = successor
            else:
                imbalanced_node = successor.parent
                self._transplant(successor, successor.right)
                successor.right = node.right
            self._transplant(node, successor)
            successor.left, successor.left.parent = node.left, successor
        self._rebalance(imbalanced_node)

    def _rebalance(self, node: _t.Union[_Nil, Node[_Key, _Value]]) -> None:
        while node is not None:
            _update_height(node)
            if node.balance_factor > 1:
                assert node.left is not NIL
                if node.left.balance_factor < 0:
                    self._rotate_left(node.left)
                self._rotate_right(node)
            elif node.balance_factor < -1:
                assert node.right is not NIL
                if node.right.balance_factor > 0:
                    self._rotate_right(node.right)
                self._rotate_left(node)
            node = node.parent

    def _rotate_left(self, node: Node[_Key, _Value]) -> None:
        replacement = node.right
        assert replacement is not NIL
        self._transplant(node, replacement)
        node.right, replacement.left = replacement.left, node
        _update_height(node)
        _update_height(replacement)

    def _rotate_right(self, node: Node[_Key, _Value]) -> None:
        replacement = node.left
        assert replacement is not NIL
        self._transplant(node, replacement)
        node.left, replacement.right = replacement.right, node
        _update_height(node)
        _update_height(replacement)

    def _transplant(self,
                    origin: Node[_Key, _Value],
                    replacement: _t.Union[_Nil, Node[_Key, _Value]]) -> None:
        parent = origin.parent
        if parent is None:
            self.root = replacement
            _set_parent(replacement, None)
        elif origin is parent.left:
            parent.left = replacement
        else:
            parent.right = replacement

    __slots__ = 'root',

    def __init__(self, root: _t.Union[_Nil, Node[_Key, _Value]]) -> None:
        self.root = root


def map_(*items: _Item[_Key, _Value]) -> _Map[_Key, _Value]:
    return _Map(Tree.from_components(*_split_items(items)))


@_t.overload
def set_(*values: _Value,
         key: None = ...) -> _Set[_Value]:
    ...


@_t.overload
def set_(*values: _Value,
         key: _Order[_Value, _Key]) -> _KeyedSet[_Key, _Value]:
    ...


def set_(
        *values: _Value,
        key: _t.Optional[_Order[_Value, _Key]] = None
) -> _t.Union[_KeyedSet[_Key, _Value], _Set[_Value]]:
    return (_Set(Tree.from_components(values))
            if key is None
            else _KeyedSet(Tree.from_components([key(value)
                                                 for value in values],
                                                values),
                           key))



from __future__ import annotations

import typing as _t
from abc import (ABC,
                 abstractmethod)
from copy import deepcopy
from itertools import chain

import typing_extensions as _te
from reprit.base import generate_repr

from .hints import (Item,
                    Key,
                    Key_co,
                    Value)
from .utils import capacity

Nil = type(None)
NIL: Nil = None


class Node(_te.Protocol[Key_co, Value]):
    @property
    @abstractmethod
    def left(self) -> _t.Union[Nil, _te.Self]:
        """Left child."""

    @left.setter
    @abstractmethod
    def left(self, _value: _t.Union[Nil, _te.Self]) -> None:
        """Sets left child."""

    @property
    @abstractmethod
    def right(self) -> _t.Union[Nil, _te.Self]:
        """Right child."""

    @right.setter
    @abstractmethod
    def right(self, _value: _t.Union[Nil, _te.Self]) -> None:
        """Sets right child."""

    @property
    def item(self) -> Item[Key_co, Value]:
        return self.key, self.value

    @property
    @abstractmethod
    def key(self) -> Key_co:
        """Comparisons key."""

    @property
    @abstractmethod
    def value(self) -> Value:
        """Underlying value."""

    @value.setter
    def value(self, value: Value) -> None:
        pass


class Tree(ABC, _t.Generic[Key, Value]):
    root: _t.Union[Nil, Node[Key, Value]]

    __slots__ = 'root',

    def __init__(self, root: _t.Union[Nil, Node[Key, Value]]) -> None:
        self.root = root

    def __bool__(self) -> bool:
        """Checks if the tree has nodes."""
        return self.root is not NIL

    def __copy__(self) -> _te.Self:
        return type(self)(deepcopy(self.root))

    def __iter__(self) -> _t.Iterator[Node[Key, Value]]:
        """Returns iterator over nodes in ascending keys order."""
        node = self.root
        queue = []
        while True:
            while node is not NIL:
                queue.append(node)
                node = node.left
            if not queue:
                return
            node = queue.pop()
            yield node
            node = node.right

    def __len__(self) -> int:
        """Returns number of nodes."""
        return capacity(self)

    def __reversed__(self) -> _t.Iterator[Node[Key, Value]]:
        """Returns iterator over nodes in descending keys order."""
        node = self.root
        queue = []
        while True:
            while node is not NIL:
                queue.append(node)
                node = node.right
            if not queue:
                return
            node = queue.pop()
            yield node
            node = node.left

    @_t.overload
    @classmethod
    def from_components(cls,
                        keys: _t.Iterable[Key],
                        values: None = ...) -> Tree[Key, Key]:
        ...

    @_t.overload
    @classmethod
    def from_components(cls,
                        keys: _t.Iterable[Key],
                        values: _t.Iterable[Value]) -> _te.Self:
        ...

    @classmethod
    @abstractmethod
    def from_components(
            cls: _t.Union[_t.Type[Tree[Key, Key]], _t.Type[Tree[Key, Value]]],
            keys: _t.Iterable[Key],
            values: _t.Optional[_t.Iterable[Value]] = None
    ) -> _t.Union[Tree[Key, Key], Tree[Key, Value]]:
        """Constructs tree from given components."""

    __repr__ = generate_repr(from_components,
                             with_module_name=True)

    @property
    def keys(self) -> _t.List[Key]:
        return _t.cast(_t.List[Key], [node.key for node in self])

    @property
    def values(self) -> _t.List[Value]:
        return [node.value for node in self]

    def clear(self) -> None:
        self.root = NIL

    def find(self, key: Key) -> _t.Union[Nil, Node[Key, Value]]:
        """Searches for the node corresponding to a key."""
        node = self.root
        while node is not NIL:
            if key < node.key:
                node = node.left
            elif node.key < key:
                node = node.right
            else:
                break
        return node

    def infimum(self, key: Key) -> _t.Union[Nil, Node[Key, Value]]:
        """Returns first node with a key not greater than the given one."""
        node, result = self.root, NIL
        while node is not NIL:
            if key < node.key:
                node = node.left
            elif node.key < key:
                result, node = node, node.right
            else:
                result = node
                break
        return result

    @abstractmethod
    def insert(self, key: Key, value: Value) -> Node[Key, Value]:
        """Inserts given key-value pair in the tree."""

    def max(self) -> _t.Union[Nil, Node[Key, Value]]:
        """Returns node with the maximum key."""
        node = self.root
        if node is not NIL:
            while node.right is not NIL:
                node = node.right
        return node

    def min(self) -> _t.Union[Nil, Node[Key, Value]]:
        """Returns node with the minimum key."""
        node = self.root
        if node is not NIL:
            while node.left is not NIL:
                node = node.left
        return node

    def pop(self, key: Key) -> _t.Union[Nil, Node[Key, Value]]:
        """Removes node with given key from the tree."""
        node = self.find(key)
        if node is not NIL:
            self.remove(node)
        return node

    def popmin(self) -> _t.Union[Nil, Node[Key, Value]]:
        node = self.root
        if node is not NIL:
            while node.left is not NIL:
                node = node.left
            self.remove(node)
        return node

    def popmax(self) -> _t.Union[Nil, Node[Key, Value]]:
        node = self.root
        if node is not NIL:
            while node.right is not NIL:
                node = node.right
            self.remove(node)
        return node

    @abstractmethod
    def predecessor(self,
                    node: Node[Key, Value]) -> _t.Union[Nil, Node[Key, Value]]:
        """Returns last node with a key less than of the given one."""

    @abstractmethod
    def remove(self, node: Node[Key, Value]) -> None:
        """Removes node from the tree."""

    @abstractmethod
    def successor(self,
                  node: Node[Key, Value]) -> _t.Union[Nil, Node[Key, Value]]:
        """Returns first node with a key greater than of the given one."""

    def supremum(self, key: Key) -> _t.Union[Nil, Node[Key, Value]]:
        """Returns first node with a key not less than the given one."""
        node, result = self.root, NIL
        while node is not NIL:
            if key < node.key:
                result, node = node, node.left
            elif node.key < key:
                node = node.right
            else:
                result = node
                break
        return result


class AbstractSet(ABC, _t.Generic[Value]):
    def __and__(self, other: AbstractSet[Value]) -> _te.Self:
        """Returns intersection of the set with given one."""
        return (self.from_iterable(value for value in self if value in other)
                if isinstance(other, AbstractSet)
                else NotImplemented)

    @abstractmethod
    def __contains__(self, value: Value) -> bool:
        """Checks if given value is presented in the set."""

    @_t.overload
    def __eq__(self, other: _te.Self) -> bool:
        ...

    @_t.overload
    def __eq__(self, other: _t.Any) -> _t.Any:
        ...

    def __eq__(self, other: _t.Any) -> _t.Any:
        """Checks if the set is equal to given one."""
        return (len(self) == len(other) and self <= other <= self
                if isinstance(other, AbstractSet)
                else NotImplemented)

    @_t.overload
    def __ge__(self, other: _te.Self) -> bool:
        ...

    @_t.overload
    def __ge__(self, other: _t.Any) -> _t.Any:
        ...

    def __ge__(self, other: _t.Any) -> _t.Any:
        """Checks if the set is a superset of given one."""
        return (len(self) >= len(other)
                and all(value in self for value in other)
                if isinstance(other, AbstractSet)
                else NotImplemented)

    @_t.overload
    def __gt__(self, other: _te.Self) -> bool:
        ...

    @_t.overload
    def __gt__(self, other: _t.Any) -> _t.Any:
        ...

    def __gt__(self, other: _t.Any) -> _t.Any:
        """Checks if the set is a strict superset of given one."""
        return (len(self) > len(other)
                and self >= other and self != other
                if isinstance(other, AbstractSet)
                else NotImplemented)

    @abstractmethod
    def __iter__(self) -> _t.Iterator[Value]:
        """Returns iterator over the set values."""

    @_t.overload
    def __le__(self, other: _te.Self) -> bool:
        ...

    @_t.overload
    def __le__(self, other: _t.Any) -> _t.Any:
        ...

    def __le__(self, other: _t.Any) -> _t.Any:
        """Checks if the set is a subset of given one."""
        return (len(self) <= len(other)
                and all(value in other for value in self)
                if isinstance(other, AbstractSet)
                else NotImplemented)

    @abstractmethod
    def __len__(self) -> int:
        """Returns size of the set."""

    @_t.overload
    def __lt__(self, other: _te.Self) -> bool:
        ...

    @_t.overload
    def __lt__(self, other: _t.Any) -> _t.Any:
        ...

    def __lt__(self, other: _t.Any) -> _t.Any:
        """Checks if the set is a strict subset of given one."""
        return (len(self) < len(other)
                and self <= other and self != other
                if isinstance(other, AbstractSet)
                else NotImplemented)

    def __or__(self, other: AbstractSet[Value]) -> _te.Self:
        """Returns union of the set with given one."""
        return (self.from_iterable(chain(self, other))
                if isinstance(other, AbstractSet)
                else NotImplemented)

    def __sub__(self, other: AbstractSet[Value]) -> _te.Self:
        """Returns subtraction of the set with given one."""
        return (self.from_iterable(value
                                   for value in self
                                   if value not in other)
                if isinstance(other, AbstractSet)
                else NotImplemented)

    def __xor__(self, other: AbstractSet[Value]) -> _te.Self:
        """Returns symmetric difference of the set with given one."""
        return ((self - other) | (other - self)
                if isinstance(other, AbstractSet)
                else NotImplemented)

    @abstractmethod
    def from_iterable(self, _value: _t.Iterable[Value]) -> _te.Self:
        """Constructs set from given values."""

    def isdisjoint(self, other: _te.Self) -> bool:
        """Checks if the tree has no intersection with given one."""
        return (all(value not in other for value in self)
                if len(self) < len(other)
                else all(value not in self for value in other))


class MutableSet(AbstractSet[Value]):
    def __iand__(self, other: AbstractSet[Value]) -> _te.Self:
        """Intersects the set with given one in-place."""
        if not isinstance(other, AbstractSet):
            return NotImplemented
        for value in self - other:
            self.discard(value)
        return self

    def __ior__(self, other: AbstractSet[Value]) -> _te.Self:
        """Unites the set with given one in-place."""
        if not isinstance(other, AbstractSet):
            return NotImplemented
        for value in other:
            self.add(value)
        return self

    def __isub__(self, other: AbstractSet[Value]) -> _te.Self:
        """Subtracts from the set a given one in-place."""
        if not isinstance(other, AbstractSet):
            return NotImplemented
        if self == other:
            self.clear()
        else:
            for value in other:
                self.discard(value)
        return self

    def __ixor__(self, other: AbstractSet[Value]) -> _te.Self:
        """Exclusively disjoins the set with given one in-place."""
        if not isinstance(other, AbstractSet):
            return NotImplemented
        if self == other:
            self.clear()
        else:
            for value in other:
                if value in self:
                    self.discard(value)
                else:
                    self.add(value)
        return self

    @abstractmethod
    def add(self, value: Value) -> None:
        """Adds given value to the set."""

    @abstractmethod
    def clear(self) -> None:
        """Adds given value to the set."""

    @abstractmethod
    def discard(self, value: Value) -> None:
        """Removes given value from the set if it is present."""

    @abstractmethod
    def remove(self, value: Value) -> None:
        """Removes given value that is present in the set."""



from __future__ import annotations

import typing as _t
from collections import abc

from reprit.base import generate_repr

from .abcs import (NIL,
                   AbstractSet,
                   Node,
                   Tree)
from .hints import (Item,
                    Key,
                    Value)
from .utils import split_items


class BaseView(_t.Generic[Key, Value]):
    __slots__ = 'tree',

    def __init__(self, tree: Tree[Key, Value]) -> None:
        self.tree = tree

    __repr__ = generate_repr(__init__)

    def __len__(self) -> int:
        return len(self.tree)


@abc.Set.register
class ItemsView(BaseView[Key, Value], AbstractSet[Item[Key, Value]]):
    def __contains__(self, item: Item[Key, Value]) -> bool:
        key, value = item
        node = self.tree.find(key)
        return node is not NIL and node.value == value

    def __iter__(self) -> _t.Iterator[Item[Key, Value]]:
        for node in self.tree:
            yield node.item

    def __reversed__(self) -> _t.Iterator[Item[Key, Value]]:
        for node in reversed(self.tree):
            yield node.item

    def from_iterable(
            self, _value: _t.Iterable[Item[Key, Value]]
    ) -> ItemsView[Key, Value]:
        keys, values = split_items(list(_value))
        return ItemsView(self.tree.from_components(keys, values))


@abc.Set.register
class KeysView(BaseView[Key, _t.Any], AbstractSet[Key]):
    def __contains__(self, key: Key) -> bool:
        return self.tree.find(key) is not NIL

    def __iter__(self) -> _t.Iterator[Key]:
        for node in self.tree:
            yield node.key

    def __reversed__(self) -> _t.Iterator[Key]:
        for node in reversed(self.tree):
            yield node.key

    def from_iterable(self, _value: _t.Iterable[Key]) -> KeysView[Key]:
        return KeysView(self.tree.from_components(_value))


class ValuesView(BaseView[Key, Value]):
    def __contains__(self, value: Value) -> bool:
        return any(candidate == value for candidate in self)

    def __iter__(self) -> _t.Iterator[Value]:
        for node in self.tree:
            yield node.value

    def __reversed__(self) -> _t.Iterator[Value]:
        for node in reversed(self.tree):
            yield node.value



from ._core import (hints as _hints,
                    maps as _maps,
                    sets as _sets)
from ._core.hints import (Key,
                          Value)

Item = _hints.Item[Key, Value]
Map = _maps.Map[Key, Value]
Order = _hints.Order[Value, Key]
KeyedSet = _sets.KeyedSet[Key, Value]
Set = _sets.Set[Value]



from __future__ import annotations

import typing as _t

import typing_extensions as _te

from ._core.abcs import (NIL,
                         Nil as _Nil,
                         Node as _Node,
                         Tree as _Tree)
from ._core.hints import (Item as _Item,
                          Key as _Key,
                          Order as _Order,
                          Value as _Value)
from ._core.maps import Map as _Map
from ._core.sets import (KeyedSet as _KeyedSet,
                         Set as _Set)
from ._core.utils import (split_items as _split_items,
                          to_unique_sorted_items as _to_unique_sorted_items,
                          to_unique_sorted_values as _to_unique_sorted_values)
from .binary import Node


class Tree(_Tree[_Key, _Value]):
    _header: Node[_Key, _Value]
    root: _t.Optional[Node[_Key, _Value]]

    __slots__ = '_header', 'root'

    def __init__(self,
                 root: _t.Union[_Nil, Node[_Key, _Value]]) -> None:
        self.root = root
        self._header = Node(NotImplemented, NotImplemented)

    def __iter__(self) -> _t.Iterator[Node[_Key, _Value]]:
        # we are collecting all values at once
        # because tree can be implicitly changed during iteration
        # (e.g. by simple lookup)
        # and cause infinite loops
        return _t.cast(_t.Iterator[Node[_Key, _Value]],
                       iter(list(super().__iter__())))

    def __reversed__(self) -> _t.Iterator[Node[_Key, _Value]]:
        # we are collecting all values at once
        # because tree can be implicitly changed during iteration
        # (e.g. by simple lookup)
        # and cause infinite loops
        return _t.cast(_t.Iterator[Node[_Key, _Value]],
                       iter(list(super().__reversed__())))

    @_t.overload
    @classmethod
    def from_components(cls: _t.Type[Tree[_Key, _Key]],
                        _keys: _t.Iterable[_Key],
                        _values: None = ...) -> Tree[_Key, _Key]:
        ...

    @_t.overload
    @classmethod
    def from_components(cls,
                        _keys: _t.Iterable[_Key],
                        _values: _t.Iterable[_Value]) -> _te.Self:
        ...

    @classmethod
    def from_components(
            cls: _t.Union[
                _t.Type[Tree[_Key, _Key]], _t.Type[Tree[_Key, _Value]]
            ],
            _keys: _t.Iterable[_Key],
            _values: _t.Optional[_t.Iterable[_Value]] = None
    ) -> _t.Union[Tree[_Key, _Key], Tree[_Key, _Value]]:
        keys = list(_keys)
        if not keys:
            return cls(NIL)
        elif _values is None:
            keys = _to_unique_sorted_values(keys)

            def to_simple_node(
                    start_index: int,
                    end_index: int,
                    constructor: _t.Callable[..., Node[_Key, _Key]]
                    = Node.from_simple
            ) -> Node[_Key, _Key]:
                middle_index = (start_index + end_index) // 2
                return constructor(keys[middle_index],
                                   (to_simple_node(start_index, middle_index)
                                    if middle_index > start_index
                                    else NIL),
                                   (to_simple_node(middle_index + 1, end_index)
                                    if middle_index < end_index - 1
                                    else NIL))

            return _t.cast(_t.Type[Tree[_Key, _Key]], cls)(
                    to_simple_node(0, len(keys))
            )
        else:
            items = _to_unique_sorted_items(keys, tuple(_values))

            def to_complex_node(
                    start_index: int,
                    end_index: int,
                    constructor: _t.Callable[..., Node[_Key, _Value]]
                    = Node[_Key, _Value]
            ) -> Node[_Key, _Value]:
                middle_index = (start_index + end_index) // 2
                return constructor(*items[middle_index],
                                   (to_complex_node(start_index, middle_index)
                                    if middle_index > start_index
                                    else NIL),
                                   (to_complex_node(middle_index + 1,
                                                    end_index)
                                    if middle_index < end_index - 1
                                    else NIL))

            return _t.cast(_t.Type[Tree[_Key, _Value]], cls)(
                    to_complex_node(0, len(items))
            )

    def find(self, key: _Key) -> _t.Union[_Nil, Node[_Key, _Value]]:
        if self.root is NIL:
            return NIL
        self._splay(key)
        root = self.root
        return NIL if key < root.key or root.key < key else root

    def insert(self, key: _Key, value: _Value) -> _Node[_Key, _Value]:
        if self.root is NIL:
            node = self.root = Node(key, value)
            return node
        self._splay(key)
        if key < self.root.key:
            self.root.left, self.root = NIL, Node(key, value,
                                                  self.root.left,
                                                  self.root)
        elif self.root.key < key:
            self.root.right, self.root = NIL, Node(key, value,
                                                   self.root,
                                                   self.root.right)
        return self.root

    def max(self) -> _t.Union[_Nil, _Node[_Key, _Value]]:
        node = self.root
        if node is not NIL:
            while node.right is not NIL:
                node = node.right
                assert node is not NIL
            self._splay(node.key)
        return node

    def min(self) -> _t.Union[_Nil, _Node[_Key, _Value]]:
        node = self.root
        if node is not NIL:
            while node.left is not NIL:
                node = node.left
                assert node is not NIL
            self._splay(node.key)
        return node

    def popmax(self) -> _t.Union[_Nil, _Node[_Key, _Value]]:
        if self.root is NIL:
            return self.root
        result = self.max()
        self._remove_root()
        return result

    def popmin(self) -> _t.Union[_Nil, _Node[_Key, _Value]]:
        if self.root is NIL:
            return self.root
        result = self.min()
        self._remove_root()
        return result

    def predecessor(
            self, node: _Node[_Key, _Value]
    ) -> _t.Union[_Nil, Node[_Key, _Value]]:
        assert isinstance(node, Node)
        if node.left is NIL:
            result, cursor, key = NIL, self.root, node.key
            while cursor is not node:
                assert cursor is not NIL
                if cursor.key < key:
                    result, cursor = cursor, cursor.right
                else:
                    cursor = cursor.left
        else:
            result = node.left
            assert result is not NIL
            while result.right is not NIL:
                result = result.right
                assert result is not NIL
        if result is not NIL:
            self._splay(result.key)
        return result

    def remove(self, node: _Node[_Key, _Value]) -> None:
        self._splay(node.key)
        self._remove_root()

    def successor(
            self, node: _Node[_Key, _Value]
    ) -> _t.Union[_Nil, Node[_Key, _Value]]:
        assert isinstance(node, Node)
        if node.right is NIL:
            result, cursor, key = NIL, self.root, node.key
            while cursor is not node:
                assert cursor is not NIL
                if key < cursor.key:
                    result, cursor = cursor, cursor.left
                else:
                    cursor = cursor.right
        else:
            result = node.right
            assert result is not NIL
            while result.left is not NIL:
                result = result.left
                assert result is not NIL
        if result is not NIL:
            self._splay(result.key)
        return result

    def _splay(self, key: _Key) -> None:
        next_root = self.root
        next_root_left_child = next_root_right_child = self._header
        while True:
            assert next_root is not NIL
            if key < next_root.key:
                if next_root.left is NIL:
                    break
                elif key < next_root.left.key:
                    next_root = self._rotate_right(next_root)
                    if next_root.left is NIL:
                        break
                next_root_right_child.left = next_root
                next_root_right_child, next_root = next_root, next_root.left
            elif next_root.key < key:
                if next_root.right is NIL:
                    break
                elif next_root.right.key < key:
                    next_root = self._rotate_left(next_root)
                    if next_root.right is NIL:
                        break
                next_root_left_child.right = next_root
                next_root_left_child, next_root = next_root, next_root.right
            else:
                break
        assert next_root is not NIL
        next_root_left_child.right, next_root_right_child.left = (
            next_root.left, next_root.right
        )
        next_root.left, next_root.right = self._header.right, self._header.left
        self.root = next_root

    def _remove_root(self) -> None:
        root = self.root
        assert root is not NIL
        if root.left is NIL:
            self.root = root.right
        else:
            right_root_child = root.right
            self.root = root.left
            self._splay(root.key)
            assert self.root is not NIL
            self.root.right = right_root_child

    @staticmethod
    def _rotate_left(
            node: Node[_Key, _Value]
    ) -> Node[_Key, _Value]:
        replacement = node.right
        assert replacement is not NIL
        node.right, replacement.left = replacement.left, node
        return replacement

    @staticmethod
    def _rotate_right(
            node: Node[_Key, _Value]
    ) -> Node[_Key, _Value]:
        replacement = node.left
        assert replacement is not NIL
        node.left, replacement.right = replacement.right, node
        return replacement


def map_(*items: _Item[_Key, _Value]) -> _Map[_Key, _Value]:
    return _Map(Tree.from_components(*_split_items(items)))


@_t.overload
def set_(*values: _Value,
         key: None = ...) -> _Set[_Value]:
    ...


@_t.overload
def set_(*values: _Value,
         key: _Order[_Value, _Key]) -> _KeyedSet[_Key, _Value]:
    ...


def set_(
        *values: _Value,
        key: _t.Optional[_Order[_Value, _Key]] = None
) -> _t.Union[_KeyedSet[_Key, _Value], _Set[_Value]]:
    return (_Set(Tree.from_components(values))
            if key is None
            else _KeyedSet(Tree.from_components([key(value)
                                                 for value in values],
                                                values),
                           key))



from __future__ import annotations

import typing as _t

import typing_extensions as _te
from reprit.base import generate_repr as _generate_repr

from ._core.abcs import (NIL,
                         Nil as _Nil,
                         Node as _Node,
                         Tree as _Tree)
from ._core.hints import (Item as _Item,
                          Key as _Key,
                          Order as _Order,
                          Value as _Value)
from ._core.maps import Map as _Map
from ._core.sets import (KeyedSet as _KeyedSet,
                         Set as _Set)
from ._core.utils import (are_keys_equal as _are_keys_equal,
                          split_items as _split_items,
                          to_unique_sorted_items as _to_unique_sorted_items,
                          to_unique_sorted_values as _to_unique_sorted_values)


class Node(_t.Generic[_Key, _Value]):
    _left: _t.Union[_Nil, _te.Self]
    _right: _t.Union[_Nil, _te.Self]

    __slots__ = '_left', '_right', '_key', '_value'

    def __init__(self,
                 key: _Key,
                 value: _Value,
                 left: _t.Union[_Nil, _te.Self] = NIL,
                 right: _t.Union[_Nil, _te.Self] = NIL) -> None:
        self._key, self._value, self._left, self._right = (key, value, left,
                                                           right)

    __repr__ = _generate_repr(__init__)

    @classmethod
    def from_simple(cls: _t.Type[Node[_Key, _Key]],
                    key: _Key,
                    *args: _t.Any) -> Node[_Key, _Key]:
        return cls(key, key, *args)

    @property
    def item(self) -> _Item[_Key, _Value]:
        return self.key, self.value

    @property
    def key(self) -> _Key:
        return self._key

    @property
    def left(self) -> _t.Union[_Nil, _te.Self]:
        return self._left

    @left.setter
    def left(self, value: _t.Union[_Nil, _te.Self]) -> None:
        self._left = value

    @property
    def right(self) -> _t.Union[_Nil, _te.Self]:
        return self._right

    @right.setter
    def right(self, value: _t.Union[_Nil, _te.Self]) -> None:
        self._right = value

    @property
    def value(self) -> _Value:
        return self._value

    @value.setter
    def value(self, value: _Value) -> None:
        self._value = value


class Tree(_Tree[_Key, _Value]):
    root: _t.Optional[Node[_Key, _Value]]

    @_t.overload
    @classmethod
    def from_components(cls,
                        _keys: _t.Iterable[_Key],
                        _values: None = ...) -> Tree[_Key, _Key]:
        ...

    @_t.overload
    @classmethod
    def from_components(cls,
                        _keys: _t.Iterable[_Key],
                        _values: _t.Iterable[_Value]) -> _te.Self:
        ...

    @classmethod
    def from_components(
            cls: _t.Union[
                _t.Type[Tree[_Key, _Key]], _t.Type[Tree[_Key, _Value]]
            ],
            _keys: _t.Iterable[_Key],
            _values: _t.Optional[_t.Iterable[_Value]] = None
    ) -> _t.Union[Tree[_Key, _Key], Tree[_Key, _Value]]:
        keys = list(_keys)
        if not keys:
            return cls(NIL)
        elif _values is None:
            keys = _to_unique_sorted_values(keys)

            def to_simple_node(
                    start_index: int,
                    end_index: int,
                    constructor: _t.Callable[..., Node[_Key, _Key]]
                    = Node.from_simple
            ) -> Node[_Key, _Key]:
                middle_index = (start_index + end_index) // 2
                return constructor(keys[middle_index],
                                   (to_simple_node(start_index, middle_index)
                                    if middle_index > start_index
                                    else NIL),
                                   (to_simple_node(middle_index + 1, end_index)
                                    if middle_index < end_index - 1
                                    else NIL))

            return _t.cast(_t.Type[Tree[_Key, _Key]], cls)(
                    to_simple_node(0, len(keys))
            )
        else:
            items = _to_unique_sorted_items(keys, tuple(_values))

            def to_complex_node(
                    start_index: int,
                    end_index: int,
                    constructor: _t.Callable[..., Node[_Key, _Value]] = Node
            ) -> Node[_Key, _Value]:
                middle_index = (start_index + end_index) // 2
                return constructor(*items[middle_index],
                                   (to_complex_node(start_index, middle_index)
                                    if middle_index > start_index
                                    else NIL),
                                   (to_complex_node(middle_index + 1,
                                                    end_index)
                                    if middle_index < end_index - 1
                                    else NIL))

            return _t.cast(_t.Type[Tree[_Key, _Value]], cls)(
                    to_complex_node(0, len(items))
            )

    def insert(self, key: _Key, value: _Value) -> Node[_Key, _Value]:
        parent = self.root
        if parent is NIL:
            node = self.root = Node(key, value)
            return node
        while True:
            if key < parent.key:
                if parent.left is NIL:
                    node = parent.left = Node(key, value)
                    return node
                else:
                    parent = parent.left
            elif parent.key < key:
                if parent.right is NIL:
                    node = parent.right = Node(key, value)
                    return node
                else:
                    parent = parent.right
            else:
                return parent

    def popmax(self) -> _t.Union[_Nil, Node[_Key, _Value]]:
        node = self.root
        if node is NIL:
            return node
        elif node.right is NIL:
            self.root = node.left
            return node
        else:
            while node.right.right is not NIL:
                node = node.right
            assert node.right is not NIL
            result, node.right = node.right, node.right.left
            return result

    def popmin(self) -> _t.Union[_Nil, Node[_Key, _Value]]:
        node = self.root
        if node is NIL:
            return node
        elif node.left is NIL:
            self.root = node.right
            return node
        else:
            while node.left.left is not NIL:
                node = node.left
            assert node.left is not NIL
            result, node.left = node.left, node.left.right
            return result

    def predecessor(
            self, node: _Node[_Key, _Value]
    ) -> _t.Union[_Nil, Node[_Key, _Value]]:
        assert isinstance(node, Node)
        if node.left is NIL:
            result, cursor, key = NIL, self.root, node.key
            while cursor is not node:
                assert cursor is not NIL
                if cursor.key < key:
                    result, cursor = cursor, cursor.right
                else:
                    cursor = cursor.left
        else:
            result = node.left
            while result.right is not NIL:
                result = result.right
        return result

    def remove(self, node: _Node[_Key, _Value]) -> None:
        assert isinstance(node, Node)
        assert self.root is not NIL
        parent, key = self.root, node.key
        if _are_keys_equal(key, parent.key):
            if parent.left is NIL:
                self.root = parent.right
            else:
                node = parent.left
                if node.right is NIL:
                    self.root, node.right = node, self.root.right
                else:
                    while node.right.right is not NIL:
                        node = node.right
                    assert node.right is not NIL
                    (
                        self.root, node.right.left, node.right.right,
                        node.right
                    ) = (node.right, self.root.left, self.root.right,
                         node.right.left)
            return
        while True:
            assert parent is not NIL
            if key < parent.key:
                # search in left subtree
                assert parent.left is not NIL
                if _are_keys_equal(key, parent.left.key):
                    # remove `parent.left`
                    cursor = parent.left.left
                    if cursor is NIL:
                        parent.left = parent.left.right
                        return
                    elif cursor.right is NIL:
                        parent.left, cursor.right = cursor, parent.left.right
                    else:
                        while cursor.right.right is not NIL:
                            cursor = cursor.right
                        assert cursor.right is not NIL
                        (
                            parent.left, cursor.right.left, cursor.right.right,
                            cursor.right
                        ) = (cursor.right, parent.left.left, parent.left.right,
                             cursor.right.left)
                    return
                else:
                    parent = parent.left
            # search in right subtree
            else:
                assert parent.right is not NIL
                if _are_keys_equal(key, parent.right.key):
                    # remove `parent.right`
                    assert parent.right is not NIL
                    cursor = parent.right.left
                    if cursor is NIL:
                        parent.right = parent.right.right
                        return
                    elif cursor.right is NIL:
                        parent.right, cursor.right = cursor, parent.right.right
                    else:
                        while cursor.right.right is not NIL:
                            cursor = cursor.right
                        assert cursor.right is not NIL
                        (
                            parent.right, cursor.right.left,
                            cursor.right.right, cursor.right
                        ) = (
                            cursor.right, parent.right.left,
                            parent.right.right, cursor.right.left
                        )
                    return
                else:
                    parent = parent.right

    def successor(
            self, node: _Node[_Key, _Value]
    ) -> _t.Union[_Nil, Node[_Key, _Value]]:
        assert isinstance(node, Node)
        if node.right is NIL:
            result, cursor, key = NIL, self.root, node.key
            while cursor is not node:
                assert cursor is not NIL
                if key < cursor.key:
                    result, cursor = cursor, cursor.left
                else:
                    cursor = cursor.right
        else:
            result = node.right
            assert result is not NIL
            while result.left is not NIL:
                result = result.left
                assert result is not NIL
        return result

    __slots__ = 'root',

    def __init__(self, root: _t.Union[_Nil, Node[_Key, _Value]]) -> None:
        self.root = root


def map_(*items: _Item[_Key, _Value]) -> _Map[_Key, _Value]:
    return _Map(Tree.from_components(*_split_items(items)))


@_t.overload
def set_(*values: _Value,
         key: None = ...) -> _Set[_Value]:
    ...


@_t.overload
def set_(*values: _Value,
         key: _Order[_Value, _Key]) -> _KeyedSet[_Key, _Value]:
    ...


def set_(
        *values: _Value,
        key: _t.Optional[_Order[_Value, _Key]] = None
) -> _t.Union[_KeyedSet[_Key, _Value], _Set[_Value]]:
    return (_Set(Tree.from_components(values))
            if key is None
            else _KeyedSet(Tree.from_components([key(value)
                                                 for value in values],
                                                values),
                           key))



from __future__ import annotations

import typing as _t
from collections import abc

import typing_extensions as _te
from reprit.base import generate_repr

from .abcs import (NIL,
                   Node,
                   Tree)
from .hints import (Item,
                    Key,
                    Value)
from .utils import split_items
from .views import (ItemsView,
                    KeysView,
                    ValuesView)


class Map(_t.Generic[Key, Value]):
    __slots__ = 'tree',

    def __init__(self, tree: Tree[Key, _t.Optional[Value]]) -> None:
        self.tree = tree

    __repr__ = generate_repr(__init__)

    def __contains__(self, key: Key) -> bool:
        return self.tree.find(key) is not NIL

    def __copy__(self) -> Map[Key, Value]:
        return Map(self.tree.__copy__())

    def __delitem__(self, key: Key) -> None:
        node = self.tree.pop(key)
        if node is NIL:
            raise KeyError(key)

    @_t.overload
    def __eq__(self, other: _te.Self) -> bool:
        ...

    @_t.overload
    def __eq__(self, other: _t.Any) -> _t.Any:
        ...

    def __eq__(self, other: _t.Any) -> _t.Any:
        return (self.keys() == other.keys()
                and all(other[key] == value
                        for key, value in self.items())
                if isinstance(other, Map)
                else NotImplemented)

    def __getitem__(self, key: Key) -> _t.Optional[Value]:
        return self._find_node(key).value

    def __iter__(self) -> _t.Iterator[Key]:
        for node in self.tree:
            yield node.key

    def __len__(self) -> int:
        return len(self.tree)

    def __reversed__(self) -> _t.Iterator[Key]:
        for node in reversed(self.tree):
            yield node.key

    def __setitem__(self, key: Key, value: Value) -> None:
        self.tree.insert(key, value).value = value

    def ceil(self, key: Key) -> _t.Optional[Value]:
        return self._ceil_node(key).value

    def ceilitem(self, key: Key) -> Item[Key, _t.Optional[Value]]:
        return self._ceil_node(key).item

    def clear(self) -> None:
        self.tree.clear()

    def floor(self, key: Key) -> _t.Optional[Value]:
        return self._floor_node(key).value

    def flooritem(self, key: Key) -> Item[Key, _t.Optional[Value]]:
        return self._floor_node(key).item

    def get(self,
            key: Key,
            default: _t.Optional[Value] = None) -> _t.Optional[Value]:
        node = self.tree.find(key)
        return default if node is NIL else node.value

    def items(self) -> ItemsView[Key, _t.Optional[Value]]:
        return ItemsView(self.tree)

    def keys(self) -> KeysView[Key]:
        return KeysView(self.tree)

    def max(self) -> _t.Optional[Value]:
        return self._max_node().value

    def maxitem(self) -> Item[Key, _t.Optional[Value]]:
        return self._max_node().item

    def min(self) -> _t.Optional[Value]:
        return self._min_node().value

    def minitem(self) -> Item[Key, _t.Optional[Value]]:
        return self._min_node().item

    def next(self, key: Key) -> _t.Optional[Value]:
        return self._next_node(key).value

    def nextitem(self, key: Key) -> Item[Key, _t.Optional[Value]]:
        return self._next_node(key).item

    __sentinel = object()

    def pop(self,
            key: Key,
            default: Value = _t.cast(Value, __sentinel)) -> _t.Optional[Value]:
        node = self.tree.pop(key)
        if node is NIL:
            if default is self.__sentinel:
                raise KeyError(key)
            return default
        return node.value

    def popmax(self) -> _t.Optional[Value]:
        return self._popmax_node().value

    def popmaxitem(self) -> Item[Key, _t.Optional[Value]]:
        return self._popmax_node().item

    def popmin(self) -> _t.Optional[Value]:
        return self._popmin_node().value

    def popminitem(self) -> Item[Key, _t.Optional[Value]]:
        return self._popmin_node().item

    popitem = popminitem

    def prev(self, key: Key) -> _t.Optional[Value]:
        return self._prev_node(key).value

    def previtem(self, key: Key) -> Item[Key, _t.Optional[Value]]:
        return self._prev_node(key).item

    def setdefault(self,
                   key: Key,
                   default: _t.Optional[Value] = None) -> _t.Optional[Value]:
        node = self.tree.find(key)
        return (self.tree.insert(key, default)
                if node is NIL
                else node).value

    def update(
            self, other: _t.Union[_te.Self, _t.Iterable[Item[Key, Value]]] = ()
    ) -> None:
        items = _t.cast(_t.Iterable[Item[Key, Value]],
                        other.items() if isinstance(other, Map) else other)
        for key, value in items:
            self[key] = value

    def values(self) -> ValuesView[Key, _t.Optional[Value]]:
        return ValuesView(self.tree)

    def _ceil_node(self, key: Key) -> Node[Key, _t.Optional[Value]]:
        node = self.tree.supremum(key)
        if node is NIL:
            raise KeyError('No key found greater than or equal to {!r}'
                           .format(key))
        return node

    def _find_node(self, key: Key) -> Node[Key, _t.Optional[Value]]:
        result = self.tree.find(key)
        if result is NIL:
            raise KeyError(key)
        return result

    def _floor_node(self, key: Key) -> Node[Key, _t.Optional[Value]]:
        result = self.tree.infimum(key)
        if result is NIL:
            raise KeyError('No key found less than or equal to {!r}'
                           .format(key))
        return result

    def _max_node(self) -> Node[Key, _t.Optional[Value]]:
        result = self.tree.max()
        if result is NIL:
            raise KeyError('Map is empty')
        return result

    def _min_node(self) -> Node[Key, _t.Optional[Value]]:
        result = self.tree.min()
        if result is NIL:
            raise KeyError('Map is empty')
        return result

    def _next_node(self, key: Key) -> Node[Key, _t.Optional[Value]]:
        result = self.tree.successor(self._find_node(key))
        if result is NIL:
            raise KeyError('Corresponds to maximum')
        return result

    def _popmax_node(self) -> Node[Key, _t.Optional[Value]]:
        result = self.tree.popmax()
        if result is NIL:
            raise KeyError('Map is empty')
        return result

    def _popmin_node(self) -> Node[Key, _t.Optional[Value]]:
        result = self.tree.popmin()
        if result is NIL:
            raise KeyError('Map is empty')
        return result

    def _prev_node(self, key: Key) -> Node[Key, _t.Optional[Value]]:
        result = self.tree.predecessor(self._find_node(key))
        if result is NIL:
            raise KeyError('Corresponds to minimum')
        return result


abc.MutableMapping.register(Map)


def map_constructor(
        tree_constructor: _t.Callable[
            [_t.Iterable[Key], _t.Iterable[Value]],
            Tree[Key, _t.Optional[Value]]
        ],
        *items: Item[Key, Value]
) -> Map[Key, Value]:
    return Map(tree_constructor(*split_items(items)))
